package BDSM::Agent;

use strict;
use warnings;
use include;

use BDSM::Vector;

use StdLib;

my $agent = create("Agent",
  Symbol => "@"
);

$agent->{Init} = sub {
  my ($self, $opts) = @_;
  $self->_construct($opts => "Map", "Symbol", "Color");
  $self->{Void} = 1;  # Till we move somewhere
  $self->{Sprite} = $self->{Aggregate} = $self;   # We being bullied? Talk to us, then.

  # Going somewhere already?
  if ($self->{Map}) {
    $opts->{At} ? $self->go(@{ delete $opts->{At} }) : $self->warp;
    $self->{Void} = !defined $self->{Y};   # go() might've failed, mind you
  }

  return;
};

### methods

# Can we move there?
$agent->{Methods}{_preblit} = sub {
  my ($self, $offy, $offx, $interact) = @_;
  my ($y, $x) = ($offy, $offx);
  unless ($self->{Void}) {
    $y += $self->{Y};
    $x += $self->{X};
  }

  return unless $self->{Map}->stfucheck($y, $x);  # Valid coordinates?

  my $tile = $self->{Map}->get($y, $x);
  return if $tile eq "#";
  # Something's there... collide with them, maybe.
  if (ref $tile) {
    return unless $interact;
    # If we're not in a Blob and collide with a Blob, push entire Blob
    # If we push something in a Blob and we're in it too, just push the thing
    # If our Aggregate isn't us, we're in a Blob.
    if ($self->{Aggregate} == $self) {
      return $self->collide($tile->{Aggregate}, $y, $x);
    } else {
      return $self->collide($tile->{Sprite}, $y, $x);
    }
  }

  return 1;
};

# Erase ourselves from the map.
$agent->{Methods}{_blitoff} = sub {
  my $self = shift;
  return if $self->{Void};
  delete $self->{Map}->tile($self->{Y}, $self->{X})->{Actor};
  $self->{Map}->modded($self->{Y}, $self->{X});
};

# Draw ourselves onto the map.
$agent->{Methods}{_bliton} = sub {
  my $self = shift;
  $self->{Map}->tile($self->{Y}, $self->{X})->{Actor} = $self;
  $self->{Map}->modded($self->{Y}, $self->{X});
};

# If we're omnomnomed by a blob, we need to report our unicellularnicity.
$agent->{Methods}{height} = sub {
  return 0;
};

$agent->{Methods}{width} = sub {
  return 0;
};

# Return what we're standing on.
$agent->{Methods}{tile} = sub {
  my $self = shift;
  return $self->{Map}->tile($self->{Y}, $self->{X});
};

# Do we submit to somebody pushing us?
$agent->{Methods}{pushedby} = sub {
  my ($self, $pusher) = @_;
  return 1;   # Override me!
};

### go

$agent->react("Before", "go",
  role => "check",
  by   => sub {
    my ($self, $heap) = @_;
    my ($y, $x) = @{ $heap->{Args} };

    # 1) We have to get off our tile... if we can.
    if ($self->{Void} == 0 and $self->tile->{OnExit}) {
      my @code = $self->tile->{OnExit}->($self);
      return @code unless @code == 1;   # Handle the return code; this is an abnormal event.
    }

    # 2) Now check out where we're trying to go...
    my $interact;
    if ($self->{Void}) {
      $interact = 0;  # If we're nowhere, don't push stuff.
    } else {
      $y -= $self->{Y};
      $x -= $self->{X};
      $interact = 1;
    }
    return STOP unless $self->_preblit($y, $x, $interact);
  }
);

$agent->react("On", "go",
  role => "do",
  by   => sub {
    my ($self, $heap) = @_;
    my ($y, $x) = @{ $heap->{Args} };

    # 1) Blit off old position.
    $self->_blitoff;

    ($self->{Y}, $self->{X}) = ($y, $x);

    # 2) Blit onto new position.
    $self->_bliton;

    # 3) And draw changes.
    $self->{Map}->callback("draw" => $self);
  }
);

$agent->react("After", "go",
  role => "customtile",
  by   => sub {
    my ($self, $heap) = @_;
    my ($y, $x) = @{ $heap->{Args} };

    # Let the new tile react to us.
    return $self->tile->{OnEnter}->($self) if $self->tile->{OnEnter};
  }
);

### warp

$agent->react("On", "warp",
  role => "do",
  by   => sub {
    my $self = shift;
    # Explicitly blitoff now, since Void being on will stop it happening. It'd be nice if we
    # could tell go() to not interact with a flag.
    $self->_blitoff;
    $self->{Void} = 1;
    $self->go( $self->{Map}->getpt(".") );
    $self->{Void} = 0;
  }
);

### directionals

$agent->react("Before", "n",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", north_of($self));
  }
);

$agent->react("Before", "s",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", south_of($self));
  }
);

$agent->react("Before", "w",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", west_of($self));
  }
);

$agent->react("Before", "e",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", east_of($self));
  }
);

$agent->react("Before", "nw",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", northwest_of($self));
  }
);

$agent->react("Before", "ne",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", northeast_of($self));
  }
);

$agent->react("Before", "sw",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", southwest_of($self));
  }
);

$agent->react("Before", "se",
  role => "check",
  by   => sub {
    my $self = shift;
    return REDIRECT("go", southeast_of($self));
  }
);

### collide

$agent->react("Before", "collide",
  role => "check",
  by   => sub {
    my ($self, $heap) = @_;
    my ($victim, $toy, $tox) = @{ $heap->{Args} };

    # talk to em, attack em, interact... but by default, just push!
    if ($victim->pushedby($self)) {
      # Hey, they said we could.
      return STOP unless $victim->go(
        dir_relative(
          dir_of(from => [$self->{Y}, $self->{X}], to => [$toy, $tox]),
          $victim->{Y}, $victim->{X}
        )
      );
    }
    return STOP;   # Not cool, they didn't move.
  }
);

42;
