package BDSM::Agent::Blob;

use strict;
use warnings;
use include;

use BDSM::Map;

use StdLib;

my $blob = create("Blob", From => "Sprite");

$blob->{Init} = sub {
  my ($self, $opts) = @_;

  $self->_construct($opts => "Map");
  $self->{Void} = 1;
  $self->{Agents} = [];

  # Set up our internal BDSM::Map and the Shape reference so all of Sprite's rendering
  # routines like us.
  $self->{Grid} = BDSM::Map->new(0, 0, " ");
  $self->_recomblobulate;

  # Going somewhere already?
  if ($self->{Map}) {
    $opts->{At} ? $self->go(@{ delete $opts->{At} }) : $self->warp;
    $self->{Void} = !defined $self->{Y};   # go() might've failed, mind you
  }

  return;
};

### methods

$blob->{Methods}{height} = sub {
  my $self = shift;
  return $self->{Grid}->height;
};

$blob->{Methods}{width} = sub {
  my $self = shift;
  return $self->{Grid}->width;
};

$blob->{Methods}{_recomblobulate} = sub {
  # Internally, our Grid (a BDSM::Map... literally)'s individual nodes are structured
  # differently than from a normal Sprite's Shape structure... so fix that.
  # We're VERY slow right now! Call us whenever anything on the blob changes.
  my $self = shift;
  $self->{Shape} = [];  # Always start fresh in case we've grown or shrunk
  foreach my $y (0 .. $self->height) {
    foreach my $x (0 .. $self->width) {
      $self->{Shape}[$y][$x] = $self->{Grid}->get($y, $x);
    }
  }
};

$blob->{Methods}{_prune} = sub {
  # Prune off extra junk off a BDSM::Map blob
  my (undef, $map) = (@_);

  # Prune the bottom
  BOTTOM: foreach my $y (reverse(0 .. $map->height)) {
    foreach my $x (0 .. $map->width) {
      last BOTTOM if ref $map->get($y, $x);
    }
    pop @{ $map->{Map} };
  }

  # Prune the right side
  RIGHT: foreach my $x (reverse(0 .. $map->width)) {
    foreach my $y (0 .. $map->height) {
      last RIGHT if ref $map->get($y, $x);
    }
    pop @{ $map->{Map}[$_] } foreach 0 .. $map->height;
  }

  # The top and left never end up with extra because of the way the offsets work
  return $map;
};

### omnomnom

$blob->react("Before", "omnomnom",
  role => "check",
  by   => sub {
    my ($blob, $heap) = @_;
    my ($agent, $offy, $offx) = @{ $heap->{Args} };

    # Before we mess with ANYTHING, check if the new agent can fit in the blob. Rather, the
    # blob always accomodates, but the blob's environment might not. And naturally this
    # check is only required if the blob isn't in the void.
    return if $blob->{Void};

    my @save = ($agent->{Map}, $agent->{Y}, $agent->{X}, $agent->{Void});
    $agent->{Map} = $blob->{Map};
    $agent->{Y} = $blob->{Y} + $offy;
    $agent->{X} = $blob->{X} + $offx;
    $agent->{Void} = 0;

    my $code = $agent->_preblit(0, 0);
    ($agent->{Map}, $agent->{Y}, $agent->{X}, $agent->{Void}) = @save;

    return STOP("not enuff room to omnomnom an agent") unless $code;
  }
);

$blob->react("On", "omnomnom",
  role => "do",
  by   => sub {
    my ($blob, $heap) = @_;
    my ($agent, $offy, $offx) = @{ $heap->{Args} };

    # How big will our new blob be?
    # The calculations got a little... messy, so just make it large enough to work for sure
    # then prune the extra later.
    my $height = $blob->height + $agent->height + abs($offy);
    my $width = $blob->width + $agent->width + abs($offx);

    # Form the blank shape
    my $shape = BDSM::Map->new($height, $width, " ");

    # Add an offset to old agents' position since we've potentially resized the blob.
    my $oldy = $offy > 0 ? 0 : abs($offy);
    my $oldx = $offx > 0 ? 0 : abs($offx);
    foreach my $old (@{ $blob->{Agents} }) {
      $old->{Map} = $shape;
      $old->{Y} += $oldy;
      $old->{X} += $oldx;
    }

    # And set up the new agent with its position
    # (Remove them from their previous boring BDSM::Map)
    $agent->_blitoff unless $agent->{Void};
    $agent->{Aggregate} = $blob;
    $agent->{Map} = $shape;
    $agent->{Y} = $offy > 0 ? $offy : 0;
    $agent->{X} = $offx > 0 ? $offx : 0;
    $agent->{Void} = 0;
    push @{ $blob->{Agents} }, $agent;

    # Merge in all agents, which we conveniently have a list of.
    $_->_bliton foreach @{ $blob->{Agents} };

    # If the agent is a sprite, we need to set its Aggregate to us... because we have
    # assimilated it... *burp*
    if ($agent->{Shape}) {
      foreach my $y (0 .. $shape->height) {
        foreach my $x (0 .. $shape->width) {
          next unless ref $shape->get($y, $x);
          $shape->get($y, $x)->{Aggregate} = $blob;
        }
      }
    }

    # Prune the extra map stuff off
    $blob->_prune($shape);

    # Render the new blob.
    if ($blob->{Void}) {
      # We're still in the void? At least save our grid. :P
      $blob->{Grid} = $shape;
      $blob->_recomblobulate;
    } else {
      $blob->_blitoff;
      $blob->{Grid} = $shape;
      $blob->{Y} += $offy if $offy < 0;
      $blob->{X} += $offx if $offx < 0;
      $blob->_recomblobulate;
      $blob->_bliton;
    }

    # Oh yeah, and add a callback to the map so that when stuff happens inside it, stuff
    # happens wherever the blob is too...
    $blob->{Grid}->hook(sub {
      my $map = shift;
      my $cmd = shift;
      my $blob = $map->{Heap}{Blob};
      if ($cmd eq "mod") {
        # TODO: uhhh do we.. handle this? i wrote it for some reason..
      } elsif ($cmd eq "draw") {
        $blob->_blitoff;
        $blob->_recomblobulate;
        $blob->_bliton;
      }
    }, Blob => $blob);
    
    # It's been a pleasure to have you.
  }
);

42;
