package BDSM::Agent::Sprite;

use strict;
use warnings;
use include;

use BDSM::Vector;

use StdLib;

my $sprite = create("Sprite", From => "Agent");

$sprite->{Init} = sub {
  my ($self, $opts) = @_;
  $self->_construct($opts => "Map", "Color");
  $self->{Void} = 1;  # Till we move somewhere.

  # Carve our shapes!
  $self->{Shapes}{$_} = $self->_carve($opts->{Shapes}{$_}) foreach keys %{ $opts->{Shapes} };
  delete $opts->{Shapes};
  $self->{CurShape} = delete $opts->{Shape} || "default";
  $self->{Shape} = $self->{Shapes}{ $self->{CurShape} };
  die "what does this sprite look like, hmm?" unless $self->{Shape};

  # Going somewhere already?
  if ($self->{Map}) {
    $opts->{At} ? $self->go(@{ delete $opts->{At} }) : $self->warp;
    $self->{Void} = !defined $self->{Y};   # go() might've failed, mind you
  }

  return;
};

### methods

$sprite->{Methods}{_carve} = sub {
  # Turn a multi-lined string into a 2D shape.
  my ($self, $string) = @_;
  my @rows = split(/\n/, $string);

  # If it's a blank row on the edge, we can ignore it.
  shift(@rows) if $rows[0] eq "" or $rows[0] =~ m/^\s+$/;
  pop(@rows) if $rows[-1] eq "" or $rows[-1] =~ m/^\s+$/;
  my $width = longest(@rows);

  # Pad if we have to
  @rows = map { $_ . " " x ($width - length($_)) } @rows;

  # Form the shape.
  my $shape = [];
  foreach my $row (@rows) {
    # We don't make piece symbols if we're just a blank space
    push @$shape, [ map { $_ eq " " ? " " : {
      Aggregate => $self,   # When you bully a piece, you bully a village...
      Sprite    => $self,   # But we still have a sense of $self. In a blob, together.
      Symbol    => $_,
      Color     => $self->{Color}
    } } split(//, $row) ];
  }

  return $shape;
};

$sprite->{Methods}{_dump} = sub {
  # Quick and simple dump to STDERR in a format suitable for viewing with most(1)
  my $self = shift;
  foreach my $row (@{ $self->{Shape} }) {
    print STDERR join("", map { ref $_ ? $_->{Symbol} : $_ } @$row) . "\n";
  }
};

$sprite->{Methods}{height} = sub {
  my $self = shift;
  return $#{ $self->{Shape} };
};

$sprite->{Methods}{width} = sub {
  my $self = shift;
  return $#{ $self->{Shape}[0] };
};

$sprite->{Methods}{_preblit} = sub {
  my ($self, $offy, $offx, $interact) = @_;

  foreach my $y (0 .. $self->height) {
    foreach my $x (0 .. $self->width) {
      next unless ref $self->{Shape}[$y][$x];

      # $offy and $offx are OFFSETS of our current position, NOT absolute coordinates
      my ($Y, $X) = ($offy, $offx);
      unless ($self->{Void}) {
        $Y += $self->{Y};
        $X += $self->{X};
      }

      return unless $self->{Map}->stfucheck($y + $Y, $x + $X);  # Valid coordinates?

      my $tile = $self->{Map}->get($y + $Y, $x + $X);
      return if $tile eq "#";

      # Something's there, but it may not even warrant a collision...
      if (ref $tile) {
        # When go() and transform() call us, the current shape may still exist on the map.
        # Since it'll disappear if there are no other conflicts, we can ignore colliding
        # with ourselves.
        next if $tile->{Aggregate} == $self;  # Entire blob
        next if $tile->{Sprite} == $self;     # One sprite in the blob

        return unless $interact;
        # If we're not in a Blob and collide with a Blob, push entire Blob
        # If we push something in a Blob and we're in it too, just push the thing
        # If our Aggregate isn't us, we're in a Blob.
        if ($self->{Aggregate} == $self) {
          return $self->collide($tile->{Aggregate}, $Y, $X);
        } else {
          return $self->collide($tile->{Sprite}, $Y, $X);
        }
      }
    }
  }
  return 1;
};

$sprite->{Methods}{_blitoff} = sub {
  my $self = shift;
  return if $self->{Void};
  foreach my $y (0 .. $self->height) {
    foreach my $x (0 .. $self->width) {
      next unless ref $self->{Shape}[$y][$x];
      delete $self->{Map}->tile($self->{Y} + $y, $self->{X} + $x)->{Actor};
      $self->{Map}->modded($self->{Y} + $y, $self->{X} + $x);
    }
  }
};

$sprite->{Methods}{_bliton} = sub {
  my $self = shift;
  foreach my $y (0 .. $self->height) {
    foreach my $x (0 .. $self->width) {
      next unless ref $self->{Shape}[$y][$x];
      $self->{Map}->tile($self->{Y} + $y, $self->{X} + $x)->{Actor} = $self->{Shape}[$y][$x];
      $self->{Map}->modded($self->{Y} + $y, $self->{X} + $x);
    }
  }
};

### go
# Override all of Agent's reactions... we're, uh, bigger than they are.

$sprite->react("Before", "go",
  role     => "check",
  override => ["Agent" => "check"],
  by       => sub {
    my ($self, $heap) = @_;
    my ($toy, $tox) = @{ $heap->{Args} };

    # This (y, x) we get is relative our top-left corner. So calculate the difference
    my ($dy, $dx) = ($toy, $tox);
    unless ($self->{Void}) {
      $dy -= $self->{Y};
      $dx -= $self->{X};
    }
    # Store these; they're useful.
    $heap->{Args} = [$dy, $dx];

    # TODO: what, check EVERY OnExit?
    
    # 1) Check out where we're trying to go...
    my $interact = !$self->{Void};  # If we're nowhere, don't push stuff.
    return STOP unless $self->_preblit($dy, $dx, $interact);
  }
);

$sprite->react("On", "go",
  role     => "do",
  override => ["Agent" => "do"],
  by       => sub {
    my ($self, $heap) = @_;
    my ($dy, $dx) = @{ $heap->{Args} };
    
    # 1) Blit off old position.
    $self->_blitoff;
    
    $self->{Y} += $dy;
    $self->{X} += $dx;
    
    # 2) Blit onto new position.
    $self->_bliton;                                                         

    # 3) And draw changes.
    $self->{Map}->callback("draw" => $self);
  }
);

$sprite->react("After", "go",
  role     => "customtile",
  override => ["Agent" => "customtile"],
  by       => sub {
    # We don't want Agent's AFTER_go cause we have many many tiles.
  }
);

### warp

$sprite->react("On", "warp",
  role     => "do",
  override => ["Agent" => "do"],
  by       => sub {
    my $self = shift;
    # Explicitly blitoff now, since Void being on will stop it happening. It'd be nice if we
    # could tell go() to not interact with a flag.
    $self->_blitoff;

    $self->{Void} = 1;
    while (1) {
      my ($y, $x) = $self->{Map}->getpt(".");
      # Inclusive due to border
      next if $y + $self->height >= $self->{Map}->height;
      next if $x + $self->width >= $self->{Map}->width;
      if ($self->_preblit($y, $x, 0)) {
        $self->go($y, $x);
        last;
      }
    }
    $self->{Void} = 0;
  }
);

### transform

$sprite->react("Before", "transform",
  role => "check",
  by   => sub {
    my ($self, $heap) = @_;
    my ($shape) = @{ $heap->{Args} };
    return STOP if $shape eq $self->{CurShape};

    # Is there room for the new shape?
    my $oldshape = $self->{Shape};
    $self->{Shape} = $self->{Shapes}{$shape};
    my $code = $self->_preblit(0, 0);   # Don't interact
    $self->{Shape} = $oldshape;
    return STOP unless $code;
  }
);

$sprite->react("On", "transform",
  role => "do",
  by   => sub {
    my ($self, $heap) = @_;
    my ($shape) = @{ $heap->{Args} };

    # 1) Disappear...
    $self->_blitoff;
    
    # 2) ... And reappear, metamorphised!
    $self->{Shape} = $self->{Shapes}{$shape};
    $self->{CurShape} = $shape;
    $self->_bliton;

    # 3) And draw changes.
    $self->{Map}->callback("draw" => $self);
  }
);

42;
