package BDSM::Map;

use strict;
use warnings;
use include;

use BDSM::Map::Level;
use BDSM::Map::Transform;
use BDSM::Map::Inspect;
# They give us methods for modifying the map.
our @ISA = ("BDSM::Map::Level", "BDSM::Map::Transform", "BDSM::Map::Inspect");

use BDSM::Agent;
use Container;

# Create a new map of specififed dimensions and $blank tiles.
sub new {
  my $class = shift;
  my ($height, $width, $blank, $load);
  if (@_ == 1) {
    $load = shift;
  } else {
    $height = shift;
    $width = shift;
    $blank = shift || " ";
  }

  my $self = bless {
    Map   => [],
    Heap  => {},
    Hooks => []
  }, $class;

  if ($load) {
    my (@ppl, %agents);
    ($self->{Map}, @ppl) = _loadmap($load);
    foreach (@ppl) {
      $agents{ $_->[0] } = GAME->{Templates}{Agent}->new(
        Map => $self,
        Symbol => $_->[0],
        At => [ $_->[1], $_->[2] ]
      );
    }
    $self->{LoadedAgents} = \%agents;
  } else {
    # Initialize the map with $blanks.
    foreach my $y (0 .. $height) {
      foreach my $x (0 .. $width) {
        $self->{Map}[$y][$x] = {
          _ => $blank
        };
      }
    }
  }

  return $self;
}

# Change a tile and perform callbacks.
sub mod {
  my ($self, $y, $x, $new) = @_;
  return unless $self->check($y, $x);
  $self->{Map}[$y][$x]{_} = $new;

  $self->modded($y, $x);
}

# Perform callbacks on a tile that has changed.
sub modded {
  my ($self, $y, $x) = @_;
  return unless $self->check($y, $x);

  $self->callback("mod", $y, $x);
}

# Adds a new hook.
sub hook {
  my ($self, $hook, $id, $ref) = @_;
  push @{ $self->{Hooks} }, $hook;
  $self->{Heap}{$id} = $ref;
}

# Create a map from a file.
sub _loadmap {
  my $file = shift;
  open my $lvl, $file or die "Can't load map $file: $!\n";

  # First just get a grid.
  my $length = 0;
  my $grid = [];
  while (my $line = <$lvl>) {
    chomp($line);
    my @row = split(//, $line);
    $length = $#row if $#row > $length;
    push @$grid, \@row;
  }
  close $lvl;

  # Pad the grid cause Thomas can't draw a map.
  foreach my $y (0 .. $#{ $grid }) {
    push @{ $grid->[$y] }, " " foreach 0 .. $length - $#{ $grid->[$y] } - 1;
  }

  # Then make it into a map with agents.
  my @agents;
  my $map = [];
  foreach my $y (0 .. $#{ $grid }) {
    foreach my $x (0 .. $#{ $grid->[0] }) {
      my $tile = $grid->[$y][$x];
      unless ($tile eq "#" or $tile eq " " or $tile eq ".") {
        push @agents, [$tile, $y, $x];
        $tile = ".";
      }
      $map->[$y][$x] = { _ => $tile };
    }
  }

  return ($map, @agents);
}

# Execute all of our hooks with an arbitrary message.
sub callback {
  my $self = shift;
  foreach (@{ $self->{Hooks} }) {
    $_->($self, @_);
  }
}

# Flyweight containers since we don't want to waste creating one on every tile of a giant
# map.
sub inv {
  my ($self, $y, $x, $cmd, @args) = @_;
  my $tile = $self->tile($y, $x);

  # If the tile has a Container, delegate.
  if ($tile->{Inv}) {
    my @code = $tile->{Inv}->$cmd(@args);
    $self->modded($y, $x);  # Re-render if necessary
    return @code;
  }

  # Handle Container's methods ourselves, sort of.
  return undef if $cmd eq "get";
  return () if $cmd eq "all";
  die "cant delete item off tile w/o a container!" if $cmd eq "del";

  # Actually create a Container.
  $tile->{Inv} = Container->new($tile);
  $tile->{Inv}->add(@args);
  $self->modded($y, $x);  # Re-render if necessary
}

42;
