package BDSM::Map::Blockmap;

use strict;
use warnings;
use include;

use BDSM::Vector;

# We extend BDSM::Map with even more methods. They don't modify the map, rather they give it
# metadata and provide a framework for collaborative emergent antiobject intelligence.
# Fancy, eh?
use BDSM::Map;
our @ISA = ("BDSM::Map");

# Interface note: Paths are IDed by the agent pursuing them.

use BDSM::Map::Blockmap::Room;
use Heap::Priority;

# Build the blockmap and set up callbacks to track agents in rooms.
sub new {
  # Go ahead and make the map.
  my $self = BDSM::Map::new(@_);

  $self->hook(sub {
    my $map = shift;
    my $cmd = shift;
    # draw means somebody's moved.
    return unless $cmd eq "draw";
    my $agent = shift;
    my $room = $agent->tile->{Room};

    # Changing rooms?
    if (!$agent->{Room} or $agent->{Room} != $room) {
      if ($agent->{Room}) {
        $agent->{Room}->agent_exit($agent);
        # And don't forget to refresh the old room.
        $agent->{Room}->find_exits;
      }
      $agent->{Room} = $room;
      $agent->{Room}->agent_enter($agent);
    }
    $agent->{Room}->find_exits;
  },
  Blockmap => {
    Blockmap => $self->_blockify
  });
  
  # The hooks to update the rooms' exits weren't in place when the agents were put in place.
  foreach (values %{ $self->{LoadedAgents} }) {
    $_->{Room} = $_->tile->{Room};
    $_->{Room}->agent_enter($_);
    $_->{Room}->find_exits;
  }

  return $self;
}

# Quick access to our heap.
sub blockdat {
  return shift()->{Heap}{Blockmap};
}

# When a room changes, mark paths involving it for recalculation.
sub exits_changed {
  my ($self, $room) = @_;
  my $data = $self->blockdat;
  foreach my $path (values %{ $data->{Rooms}{$room} }) {
    $data->{Paths}{$path}{Update} = 1;
  }
}

# Transform a potentially sparsely-populated map into a compressed blockmap.
sub _blockify {
  my $map = shift;

  # Scan for the block rows we'll need. Create a new block row when there's a change between
  # two actual rows.
  my @rows;
  my $prevrow = join("", map { $_->{_} eq "#" ? "#" : "." } @{ $map->{Map}[0] });
  my $y1 = 0;
  foreach my $y (0 .. $map->height) {
    my $row = join("", map { $_->{_} eq "#" ? "#" : "." } @{ $map->{Map}[$y] });
    if ($row ne $prevrow) {
      push @rows, {
        Y1 => $y1,
        Y2 => $y - 1
      };
      $y1 = $y;
    }
    $prevrow = $row;
  }
  push @rows, {
    Y1 => $y1,
    Y2 => $map->height
  };

  # Likewise for columns.
  my @cols;
  my $prevcol = join("", map { $_->[0]{_} eq "#" ? "#" : "." } @{ $map->{Map} });
  my $x1 = 0;
  foreach my $x (0 .. $map->width) {
    my $col = join("", map { $_->[$x]{_} eq "#" ? "#" : "." } @{ $map->{Map} });
    if ($col ne $prevcol) {
      push @cols, {
        X1 => $x1,
        X2 => $x - 1
      };
      $x1 = $x;                                                             
    }
    $prevcol = $col;
  }
  push @cols, {
    X1 => $x1,
    X2 => $map->width
  };

  # Now form the blockmap from the rows and columns.
  my $blocks = BDSM::Map->new($#rows, $#cols);
  foreach my $y (0 .. $blocks->height) {
    foreach my $x (0 .. $blocks->width) {
      my $node = $blocks->tile($y, $x);

      # Disregard agents already on the map, they aren't permanent obstacles.
      $node->{_} = $map->get($rows[$y]{Y1}, $cols[$x]{X1}) eq "#" ? "#" : ".";

      # Make a room for each corresponding tile.
      my $room = BDSM::Map::Blockmap::Room->new(
        Map    => $map,
        BlockY => $y,
        BlockX => $x,
        Y1     => $rows[$y]{Y1},
        X1     => $cols[$x]{X1},
        Y2     => $rows[$y]{Y2},
        X2     => $cols[$x]{X2}
      );
      $node->{Room} = $room;

      # Tell the real map what room it's a part of (and thus what block)
      foreach my $Y ($room->{Y1} .. $room->{Y2}) {
        foreach my $X ($room->{X1} .. $room->{X2}) {
          $map->tile($Y, $X)->{Room} = $room;
        }
      }
    }
  }

  # And finally tell each of the rooms what neighbors might be available to them.
  foreach my $y (0 .. $blocks->height) {
    foreach my $x (0 .. $blocks->width) {
      my $node = $blocks->tile($y, $x);
      next if $node->{_} eq "#";
      my $adj = { Vertical => [], Horizontal => [], Diagonal => [] };
      $node->{Room}{Neighbors} = { Vertical => [], Horizontal => [] };
      push @{ $adj->{Vertical} }, [$y - 1, $x] if $blocks->get($y - 1, $x) eq ".";
      push @{ $adj->{Vertical} }, [$y + 1, $x] if $blocks->get($y + 1, $x) eq ".";
      push @{ $adj->{Horizontal} }, [$y, $x - 1] if $blocks->get($y, $x - 1) eq ".";
      push @{ $adj->{Horizontal} }, [$y, $x + 1] if $blocks->get($y, $x + 1) eq ".";
      push @{ $adj->{Diagonal} }, [$y - 1, $x - 1] if $blocks->get($y - 1, $x - 1) eq ".";
      push @{ $adj->{Diagonal} }, [$y - 1, $x + 1] if $blocks->get($y - 1, $x + 1) eq ".";
      push @{ $adj->{Diagonal} }, [$y + 1, $x - 1] if $blocks->get($y + 1, $x - 1) eq ".";
      push @{ $adj->{Diagonal} }, [$y + 1, $x + 1] if $blocks->get($y + 1, $x + 1) eq ".";
      $node->{Room}{Neighbors} = $adj;

      # By default, all exits are available.
      $node->{Room}{Exits} = [ @{ $adj->{Vertical} }, @{ $adj->{Horizontal} }, @{ $adj->{Diagonal} } ];
    }
  }

  return $blocks;
}

# Register an agent tracking another agent.
sub track {
  my ($self, $agent, $goal, $offset) = @_;
  my $data = $self->blockdat;
  $data->{Paths}{$agent} = {
    Update => 1,
    Goal   => $goal,
    Offset => $offset,
    Path   => [],
    Rooms  => []  # List of rooms that our Path involves at any point in time
  };
  # And a reverse registration so the goal can inform its watcher of stuff.
  $data->{Goals}{$goal} ||= [];
  push @{ $data->{Goals}{$goal} }, $agent;
}

# Tell an agent the next step in pursuing its goal.
sub nextmove {
  my ($self, $src) = @_;
  my $data = $self->blockdat;
  my $path = $data->{Paths}{$src};

  my $goal = $data->{Paths}{$src}{Goal};
  my $toy = $goal->{Y} + $path->{Offset}[0];
  my $tox = $goal->{X} + $path->{Offset}[1];

  # Are we already there?
  if ($toy == $goal->{Y} and $tox == $goal->{X}) {
    # If there's no offset, obviously we can't occupy our goal agent's tile, so make sure
    # we're at least adjacent to them.
    return 42 if abs($src->{Y} - $goal->{Y}) <= 1 and abs($src->{X} - $goal->{X}) <= 1;
  } else {
    return 42 if $goal->{Y} == $src->{Y} and $goal->{X} == $src->{X};
  }

  # In the same room as our goal?
  if ($self->room($src->{Y}, $src->{X}) eq $self->room($toy, $tox)) {
    # It's naive to just go their way.
    my @mv;
    # Also acceptable to stay in place as a best move.
    foreach (adjacent_tiles("diag", $src->{Y}, $src->{X}), [$src->{Y}, $src->{X}]) {
      my ($y, $x) = @$_;
      next unless $self->get($y, $x) eq "." or ($y == $src->{Y} and $x == $src->{X});
      my $dist = euclid($y, $x, $toy, $tox);
      if ($#mv == -1 or $dist < $mv[0]) {
        @mv = ($dist, $y, $x);
      }
    }
    msg "$src->{Symbol} idling" if $mv[1] == $src->{Y} and $mv[2] == $src->{X};
    return ($mv[1], $mv[2]);
  }

  # Update the path if we need to.
  $self->roomfind($src) if $path->{Update} or $#{ $path->{Path} } == -1;

  # So we need a high-level route to our goal.
  my $toroom = $path->{Path}[0];
  my $oldrm = $self->room($src->{Y}, $src->{X});
  return unless $toroom;
  msg "#$src->{Symbol} from $oldrm->{BlockY}, $oldrm->{BlockX} to $toroom->{BlockY}, $toroom->{BlockX}";

  # Then find the best way to move towards $toroom
  my @mv;
  foreach (adjacent_tiles("diag", $src->{Y}, $src->{X})) {
    my ($y, $x) = @$_;
    next unless $self->get($y, $x) eq ".";
    # Don't get off the path.
    next unless $self->room($y, $x) eq $toroom or $self->room($y, $x) eq $oldrm;
    # The right answer is always in the direction of the old room to the new room... if that
    # node is available.
    if ($src->{Y} - $y == $oldrm->{BlockY} - $toroom->{BlockY} and $src->{X} - $x ==
      $oldrm->{BlockX} - $toroom->{BlockX}) {
      @mv = (0, $y, $x);
      last;
    }
    my $dist = euclid($y, $x, $toroom->{Y1}, $toroom->{X1});
    if ($#mv == -1 or $dist < $mv[0]) {
      @mv = ($dist, $y, $x);
    }
  }
  return ($mv[1], $mv[2]) if @mv;
  #debug "blocked on the low map level";
  return;
}

# Use A* to find the optimal block-level route for an agent to pursue its goal.
sub roomfind {
  my ($self, $src) = @_;
  my $data = $self->blockdat;
  my $path = $data->{Paths}{$src};
  $path->{Update} = 0;    # Because that's what we're doing right now

  # If we're replacing a previous path, remove our hooks into various rooms... they might
  # not be involved any longer.
  foreach my $room (@{ $path->{Rooms} }) {
    delete $self->{Rooms}{$room}{$src};
  }
  $path->{Path} = []; # In case we can't replace it, make sure we stay stuck.

  my $map       = $data->{Blockmap};
  my $goal      = $path->{Goal};
  my $startroom = $self->room($src->{Y}, $src->{X});
  my $endroom   = $self->room($goal->{Y} + $path->{Offset}[0], $goal->{X} + $path->{Offset}[1]);
  my ($y1, $x1) = ($startroom->{BlockY}, $startroom->{BlockX});
  my ($y2, $x2) = ($endroom->{BlockY}, $endroom->{BlockX});

  # Set up the priority heap for A*
  my $open = Heap::Priority->new;
  $open->lowest_first;

  # Make a temp copy of the blockmap
  my $virtmap = [];
  foreach my $y (0 .. $map->height) {
    foreach my $x (0 .. $map->width) {
      $virtmap->[$y][$x] = {};
    }
  }
  $virtmap->[$y1][$x1]{Used} = 1;

  # For backup, if the actual path is blocked, still try to take the best move to near the
  # target. Thus, store the node with the best potential closeness. But sometimes our room
  # is the best we can get, so initialize it with that possibility covered.
  my @backup;

  # From which rooms may we enter the goal room?
  my %goodend = map { ("$_->[0],$_->[1]" => 1) } $endroom->flood($goal, $src);

  my @path;

  # We start with our floodfill-found neighbors.
  foreach ($startroom->flood($src, $goal)) {
    # Hold it. We don't have to do ANYTHING if this is a one-room path...
    if ($_->[0] == $y2 and $_->[1] == $x2 and $goodend{"$startroom->{BlockY},$startroom->{BlockX}"}) {
      @path = ($map->{Map}[$y2][$x2]{Room});
    }

    my $heuristic = abs($y2 - $_->[0]) + abs($x2 - $_->[1]);
    $open->add({ Y => $_->[0], X => $_->[1], P => $heuristic }, $heuristic);
    $virtmap->[ $_->[0] ][ $_->[1] ]{Used} = 1;
  }

  #debug "$src->{Symbol} roomfinding from $y1, $x1 to $y2, $x2";

  # The main A* loop
  LOOP: while (my $node = $open->pop) {
    last if @path;  # We maybe can skip this if it's gonna be a one-room path
    # Record this node as backup if it's the closest we might get
    my $manhattan = abs($y2 - $node->{Y}) + abs($x2 - $node->{X});
    if ($#backup == -1 or $manhattan < $backup[0]) {
      @backup = ($manhattan, $node->{Y}, $node->{X});
    }

    foreach ($map->tile($node->{Y}, $node->{X})->{Room}->exits) {
      my ($y, $x) = @$_;
      next if $virtmap->[$y][$x]{Used};   # Don't retrace old trail
      $virtmap->[$y][$x]{Used} = 1;
      $virtmap->[$y][$x]{Backref} = [$node->{Y}, $node->{X}];

      # Are we at our goal?
      if ($y == $y2 and $x == $x2) {
        # Yes! But...
        # Mark it unused cause we didn't properly traverse it. There might be another way
        # in.
        $virtmap->[$y][$x]{Used} = 0;
        next unless $goodend{"$node->{Y},$node->{X}"};  # This entrance is blocked

        # We're here! Store the path.
        @path = ($map->{Map}[$y][$x]{Room});
        while (my $backref = $virtmap->[$y][$x]{Backref}) {
          ($y, $x) = @$backref;
          unshift @path, $map->{Map}[ $backref->[0] ][ $backref->[1] ]{Room};
        }

        #debug " $_->{BlockY}, $_->{BlockX}" foreach @path;
        last LOOP;
      }

      # Add it to our list.
      my $score = abs($y2 - $y) + abs($x2 - $x) + $node->{P};
      $open->add({ Y => $y, X => $x, P => $score }, $score);
    }
  }

  # Last resorts...
  if ($#path == -1) {
    if (@backup) {
      # Do the best we can.
      my ($y, $x) = ($backup[1], $backup[2]);
      @path = ($map->{Map}[$y][$x]{Room});
      while (my $backref = $virtmap->[$y][$x]{Backref}) {
        ($y, $x) = @$backref;
        unshift @path, $map->{Map}[ $backref->[0] ][ $backref->[1] ]{Room};
      }
      #debug "back) $_->{BlockY}, $_->{BlockX}" foreach @path;
    } else {
      debug "truly blocked.";
      return;
    }
  }

  $path->{Path} = \@path;
  # Register rooms on that path because we need to know if their exits change.
  $data->{Rooms}{$_}{$src} = $src foreach @path;
}

# Returns the room encompassying a tile (reference or coordinate pair)
sub room {
  my $self = shift;
  my $tile = @_ == 1 ? shift : $self->tile(@_);
  return $tile->{Room};
}

42;
