package BDSM::Map::Blockmap::Room;

use strict;
use warnings;
use include;

use BDSM::Vector;

# Create a new room on a map to encompass a range of tiles.
sub new {
  my $class = shift;
  return bless {
    Agents    => {},   # What agents are in us?
    Exits     => [],   # What exits are accessible right now?
    # Our containing map is supposed to fill out the following:
    # Map            (our container. the actual map)
    # Neighbors      (open rooms adjacent to us)
    # Y1, X1, Y2, X2 (physical coordinates in the map)
    # BlockY, BlockX (coordinates in the blockmap
    @_
  }, $class;
}

# New agent has entered us.
sub agent_enter {
  my ($self, $agent) = @_;
  $self->{Agents}{$agent} = $agent;

  my $data = $self->{Map}->blockdat;

  # If they're an active agent, update their path. Done with this room because we're
  # technically here.
  if ($data->{Paths}{$agent}) {
    shift @{ $data->{Paths}{$agent}{Path} };
  }

  # If we're a goal being tracked, let our tracker know that we've moved.
  if ($data->{Goals}{$agent}) {
    $data->{Paths}{$_}{Update} = 1 foreach @{ $data->{Goals}{$agent} };
  }
}

# Occupying agent has left us.
sub agent_exit {
  my ($self, $agent) = @_;
  delete $self->{Agents}{$agent};
}

# What adjacent rooms are accessible?
sub exits {
  my $self = shift;
  return @{ $self->{Exits} };
}

# How tall are we?
sub height {
  my $self = shift;
  return $self->{Y2} - $self->{Y1} + 1;
}

# How wide are we?
sub width {
  my $self = shift;
  return $self->{X2} - $self->{X1} + 1;
}

# Recalculate inter-room connections. Do this with some wicked cool heuristics... if a
# certain number of agents share an X coordinate, that means they might block the room
# vertically... This is WAY faster than flooding.
sub find_exits {
  my $self = shift;
  msg " reflooding $self->{BlockY}, $self->{BlockX}";

  # Keys are the x or y coordinate, value is how many agents have that coordinate.
  my (%x, %y);
  foreach my $agent (values %{ $self->{Agents} }) {
    $y{ $agent->{Y} }++;
    $x{ $agent->{X} }++;
  }

  # So if the number of agents sharing a y coordinate is >= the room's height, blocked.
  # Likewise, if those sharing an x coordinate is >= width, blocked. Haha!
  my @exits;

  if (@{ $self->{Neighbors}{Vertical} }) {
    my $clear = 1;
    foreach (values %y) {
      $clear = 0, last if $_ >= $self->width;
    }
    push @exits, @{ $self->{Neighbors}{Vertical} } if $clear;
  }

  if (@{ $self->{Neighbors}{Horizontal} }) {
    my $clear = 1;
    foreach (values %x) {
      $clear = 0, last if $_ >= $self->height;
    }
    push @exits, @{ $self->{Neighbors}{Horizontal} } if $clear;
  }

  #push @exits, @{ $self->{Neighbors}{Diagonal} };

  # Different exits available from last time?
  if (join(",", @exits) ne join(",", @{ $self->{Exits} })) {
    $self->{Exits} = \@exits;
    $self->{Map}->exits_changed($self);
  }
}

# Manually floodfill a room to find accessible exits.
sub flood {
  my ($self, $src, $accept) = @_;
  my $map = $self->{Map};
  my @nodes = ([$src->{Y}, $src->{X}]);
  my %checked = ("$src->{Y},$src->{X}" => 1);
  my (%rooms, @rooms);
  # We used to return 'values %rooms' but due to weird hash stuff, resulted in different
  # order -> nondeterminism -> different /RESULTS/ somehow.
  while (my $node = shift @nodes) {
    foreach my $adj (adjacent_tiles("diag", @$node)) {
      my ($y, $x) = @$adj;
      next if $checked{"$y,$x"};
      $checked{"$y,$x"} = 1;
      my $agent = $map->{Map}[$y][$x];
      # No agents are allowed to be in the way at all. We always flood from $src agent,
      # which is a source or goal agent. The only agent we can run into is our $acceptable
      # which is the source if we're flooding endroom and vice versa
      next unless $agent->{_} eq "." or $agent->{_} eq $accept;
      # Are we anywhere new?
      if ($agent->{Room} ne $self) {
        next if $rooms{ $agent->{Room} };
        $rooms{ $agent->{Room} } = 1;
        push @rooms, [$agent->{Room}{BlockY}, $agent->{Room}{BlockX}];
      } else {
        push @nodes, [$y, $x];
      }
    }
  }

  return @rooms;
}

42;
