package BDSM::Map::Inspect;

use strict;
use warnings;
use include;

use Heap::Priority;
use BDSM::Vector;
# We simply extend BDSM::Map with more methods. We provide info about the map.

# Verify coordinates are within bounds.
sub check {
  my ($self, $y, $x) = @_;
  if ($y < 0 or $y > $self->height or $x < 0 or $x > $self->width) {
    debug "out of bounds coords $y, $x...";
    return;
  }
  return 1;
}

# Verify coordinates are within bounds, but silently. They're expecting it to fail.
sub stfucheck {
  my ($self, $y, $x) = @_;
  return if $y < 0 or $y > $self->height or $x < 0 or $x > $self->width;
  return 1;
}

# Return the specified tile. (Full data structure)
sub tile {
  my ($self, $y, $x) = @_;
  return unless $self->check($y, $x);
  return $self->{Map}[$y][$x];
}

# Return the specified tile's residing actor or simple type.
sub get {
  my ($self, $y, $x) = @_;
  return "" unless $self->check($y, $x);  # So we can do tests w/o unitialized values
  my $tile = $self->{Map}[$y][$x];
  return $tile->{Actor} ? $tile->{Actor} : $tile->{_};
}

# Calculate map height.
sub height {
  my $self = shift;
	return $#{ $self->{Map} };
}

# Calculate map width.
sub width {
  my $self = shift;
  # All maps are rectangular; thus, each row is the same width.
	return $#{ $self->{Map}[0] };
}

# Adds a new hook.
sub hook {
  my ($self, $hook, $id, $ref) = @_;
  push @{ $self->{Hooks} }, $hook;
  $self->{Heap}{$id} = $ref;
}

# Quick and simple dump to STDERR in a format suitable for viewing with most(1)
sub _dump {
  my $self = shift;
  my ($y1, $x1, $y2, $x2);
  if (@_) {
    ($y1, $x1, $y2, $x2) = @_;
  } else {
    ($y1, $x1, $y2, $x2) = (0, 0, $self->height, $self->width);
  }
  foreach my $y ($y1 .. $y2) {
    my $line;
    foreach my $x ($x1 .. $x2) {
      my $tile = $self->{Map}[$y][$x];
      my ($sym, $color) = GAME->{Tilemap}->render($tile);
      $line .= $sym;
    }
    print STDERR "$line\n";
  }
}

# Finds a random tile on the map.
sub getpt {
  my ($self, $match) = @_;
  my ($y, $x);
  while (1) {
    $y = random(0, $self->height);
    $x = random(0, $self->width);
    next if $self->{Map}[$y][$x]{Actor};
    last if $self->{Map}[$y][$x]{_} eq $match;
  }
  return ($y, $x);
}

# A*s a path from one point to another.
sub pathfind {
  my ($map, $diag, $y1, $x1, $y2, $x2) = @_;

  # Save some time.
  return if $y1 == $x1 and $y2 == $x2;
  return unless $map->get($y2, $x2) =~ m/^(\.| )$/;

  # Set up the priority heap for A*
  my $open = Heap::Priority->new;
  $open->lowest_first;

  # Make a temp copy of the blockmap
  my $virtmap = [];
  foreach my $y (0 .. $map->height) {
    foreach my $x (0 .. $map->width) {
      $virtmap->[$y][$x] = {};
    }
  }
  $open->add({ Y => $y1, X => $x1, P => 0 }, 0);
  $virtmap->[$y1][$x1]{Used} = 1;

  while (my $node = $open->pop) {
    # Success?
    if ($node->{Y} == $y2 and $node->{X} == $x2) {
      my ($y, $x) = ($node->{Y}, $node->{X});
      my @path = ([$y, $x]);
      while (my $backref = $virtmap->[$y][$x]{Backref}) {
        ($y, $x) = @$backref;
        unshift @path, [$y, $x];
      }
      shift @path;  # We already know about ($y1, $x1).
      return @path;
    }

    foreach (adjacent_tiles($diag, $node->{Y}, $node->{X})) {
      my ($y, $x) = @$_;
      next if $virtmap->[$y][$x]{Used};
      next unless $map->get($y, $x) =~ m/^(\.| )$/;

      my $h = abs($y - $y2) + abs($x - $x2) + $node->{P};
      $virtmap->[$y][$x]{Used} = 1;
      $virtmap->[$y][$x]{Backref} = [$node->{Y}, $node->{X}];
      $open->add({ Y => $y, X => $x, P => $h }, $h);
    }
  }
  return;
}

42;
