package Hexed::Map;

use strict;
use warnings;
use include;

use StdLib::Tilemap;

use Cursed::Window;
our @ISA = ("Cursed::Window");

use Curses;

# Create and return a new menu window.
sub spawn {
  my $self = shift;

  $self->bindmap(delete $self->{Opts}{Map}) if $self->{Opts}{Map};
}

# Bind the window to render a fixed-size map.
sub bindmap {
  my ($self, $map) = @_;
  $self->{Map} = $map;
  my ($height, $width) = ($map->height, $map->width);
  $self->{OffY} = $self->{OffX} = 0;

  # Make the pad.
  $self->{Pad} = newpad($height + 1, $width + 1);
  # Fill it with the map's initial contents.
  foreach my $y (0 .. $height) {
    foreach my $x (0 .. $width) {
      my ($sym, $color) = GAME->{Tilemap}->render( $map->tile($y, $x) );
      $self->{Pad}->addch($y, $x, ord($sym) | $self->ui->paint($color));
    }
  }

  # Set up a callback.
  $map->hook(sub {
    my $map = shift;
    my $cmd = shift;
    my $win = $map->{Heap}{HexedMap};
    if ($cmd eq "mod") {
      my ($y, $x) = @_;
      my ($sym, $color) = GAME->{Tilemap}->render( $map->tile($y, $x) );
      $win->{Pad}->addch($y, $x, ord($sym) | $self->ui->paint($color));
    } elsif ($cmd eq "draw") {
      $map->{Heap}{HexedMap}->draw;
    }
  }, HexedMap => $self) unless $map->{HexHooked};   # Don't register more if we change maps
  $map->{HexHooked} = 1;
}

# Overloaded... Sort of. We're the method that actually prints to the screen. Through pads.
sub windraw {
  my $self = shift;

	$self->{Pad}->prefresh(
	  $self->{OffY}, $self->{OffX},                                     # Where in the map
	  $self->{AbsPosY} + $self->{Y1}, $self->{AbsPosX} + $self->{X1},   # Where in window
	  # TODO: potential offset issue. why is -1 needed?
	  $self->{AbsPosY} + $self->{Y2} - 1, $self->{AbsPosX} + $self->{X2} - 1
	);
}

# Just draw normally.
sub redraw {
  my $self = shift;
  $self->draw;
}

# Hey, someone could ask. It depends on our Map.
sub mapheight {
  return shift()->{Map}->height;
}

# Hey, someone could ask. It depends on our Map.
sub mapwidth {
  return shift()->{Map}->width;
}

# Shift the camera right.
sub right {
  my $self = shift;
  $self->{OffX}++ if $self->{OffX} <= $self->mapwidth - $self->{Width};
}

# Shift the camera left.
sub left {
  my $self = shift;
  $self->{OffX}-- if $self->{OffX} >= 0;
}

# Shift the camera up.
sub up {
  my $self = shift;
  $self->{OffY}-- if $self->{OffY} >= 0;
}                                                                         

# Shift the camera down.
sub down {
  my $self = shift;
  $self->{OffY}++ if $self->{OffY} <= $self->mapheight - $self->{Height};
}

# Modify the camera's offsets to center a sprite.
sub focus {
  my ($self, $sprite) = @_;

  # Dunno where the +1 offset comes from for the upper bound

  my $y = 0;
  $y = $sprite->{Y} - int($self->{Height} / 2);
  $y = $self->mapheight - $self->{Height} + 1 if $y > $self->mapheight - $self->{Height} + 1;
  $y = 0 if $y < 0;   # Lower bound here, upper above. This order is important.

  my $x = 0;
  $x = $sprite->{X} - int($self->{Width} / 2);
  $x = $self->mapwidth - $self->{Width} + 1 if $x > $self->mapwidth - $self->{Width} + 1;
  $x = 0 if $x < 0;   # Lower bound here, upper above. This order is important.

  ($self->{OffY}, $self->{OffX}) = ($y, $x);
}

# Blockingly let the user explore the current cross-section of the map with a cursor,
# performing arbitrary callbacks.
sub target {
  my $self = shift;
  my $refpt = shift;
  my $callback = shift;

  # Don't return till the user is done playing. So we act like a blocking function.
  $self->ui->start_in;

  my $map = $self->{Map};

  # Where's our cursor start? The player's value + camera's offset if necessary?
  my $miny = $self->{OffY};
  my $minx = $self->{OffX};
  # Not sure why I need that -1 offset.. prolly some stupid bug
  my $maxy = $miny + $self->height - 1;
  $maxy = $map->height if $map->height < $maxy;
  my $maxx = $minx + $self->width - 1;
  $maxx = $map->width if $map->width< $maxx;
  my $y = $refpt->{Y};
  my $x = $refpt->{X};
  $self->_highlight($y, $x);

  until ((my $key = $self->ui->wait_input) eq "<^M>") {
    $self->_unhighlight($y, $x);

    if ($key eq "<KEY_LEFT>" or $key eq "h") {
      $x-- unless $x == $minx;
    } elsif ($key eq "<KEY_RIGHT>" or $key eq "l") {
      $x++ unless $x == $maxx;
    } elsif ($key eq "<KEY_UP>" or $key eq "k") {
      $y-- unless $y == $miny;
    } elsif ($key eq "<KEY_DOWN>" or $key eq "j") {
      $y++ unless $y == $maxy;
    }

    $self->_highlight($y, $x);

    $callback->($self->{Map}, $y, $x);
  }
  $self->_unhighlight($y, $x);
  $self->draw;

  $self->ui->stop_in;

  return ($y, $x);
}

# Highlight a cell and remember its original color.
sub _highlight {
  my ($self, $y, $x) = @_;
  my $map = $self->{Map};
  my $tile = $map->tile($y, $x);

  my (undef, $color) = GAME->{Tilemap}->render($tile);

  # If it's an Object, store its instance color (which may be nothing) but grab its actual
  # color. Likewise for inventory.
  if ($tile->{Actor}) {
    $tile = $tile->{Actor};
  } elsif ($tile->{Inv} and my ($item) = $tile->{Inv}->all) {
    $tile = $item;
  }
  $tile->{_OldColor} = $tile->{Color};

  if ($color) {
    $color =~ s!/\w+!!;
    $color .= "/white";
  } else {
    $color = "grey/white";
  }
  $tile->{Color} = $color;
  $map->modded($y, $x);
  # TODO: Why a draw here but not in unhighlight?
  $self->draw;
}

# Return a cell to its original color.
sub _unhighlight {
  my ($self, $y, $x) = @_;
  my $map = $self->{Map};
  my $tile = $map->tile($y, $x);

  # Restore the old color to the right thing: actor, item, or tile.
  if ($tile->{Actor}) {
    $tile = $tile->{Actor};
  } elsif ($tile->{Inv} and my ($item) = $tile->{Inv}->all) {
    $tile = $item;
  }

  $tile->{Color} = delete $tile->{_OldColor};
  $map->modded($y, $x);
}

42;
