package HexedUI;

use strict;
use warnings;
use include;

use POE ("Wheel::Curses");
use Curses;
use Memoize;

# Extend the singleton!
use Cursed::Frame;
our @ISA = ("Cursed::Frame");

# The HexedUI singleton maintains the POE Curses session and is the lowest-level around
# terminal resizing, key input, and other events.
sub cast {
  my ($class, $layout) = @_;

  my $self = bless {
    _Keymaps  => [],
    _Shutdown => delete $layout->{Shutdown} // sub {}
  }, $class;

  # Set up the main event loop magic.
  POE::Session->create(
    inline_states => {
      _start => sub {
        $_[HEAP]{Curses} = POE::Wheel::Curses->new(
          InputEvent => "keystroke_handler"
        );
        $_[HEAP]{UI} = $self;
        $_[KERNEL]->alias_set("hexedui");
        $_[KERNEL]->sig(WINCH => "term_resize");
      },
      _stop => sub {
        # TODO: portable please
        system("clear");
        endwin;
      },
      keystroke_handler => sub {
        my $key = $_[ARG0];
        if ($key lt " ") {
          $key = "<" . uc(unctrl($key)) . ">";
        } elsif ($key =~ m/^\d{2,}$/) {
          $key = "<" . uc(keyname($key)) . ">";
        }
        $key = "<ESCAPE>" if $key eq "<^[>";  # Rather more usable.

        my $ui = $_[HEAP]{UI};
        if ($key eq "<^C>") {
          delete $_[HEAP]{curses};
          $ui->{_Shutdown}->();
          exit;
        } else {
          $ui->{_Keymaps}[-1]{Handler}->($ui->{_Keymaps}[-1]{Win}, $key);
        }
      },
      term_resize => sub {
        # Delegate to each of the windows
        my $ui = $_[HEAP]{UI};
        ### TODO: UNHANDLED!
      }
    }
  );

  # Set up curses stuff.
  curs_set(0);

  # Yo mama uses a monochrome terminal.
  my @colors = ( qw(BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE) );
  # Magically initialize all possible combos of foregrounds and backgrounds.
  my $cnt = 0;
  no strict "refs";   # Naughty!
  foreach my $bg (@colors) {
    foreach my $fg (@colors) {
      init_pair(++$cnt, &{ "COLOR_$fg" }, &{ "COLOR_$bg" });
    }
  }

  # Create the windows from the given layout.
  while (my ($win, $dat) = each %$layout) {
    $self->_make_win($win => $dat);
  }

  return $self;
}

# Any window that does input (one at a time gets it) needs to register a keymap so POE can
# call it.
sub install_keymap {
  my ($self, $win, $sub) = @_;
  # Nastyish closure magic so the handler can know who it is.
  push @{ $self->{_Keymaps} }, {Win => $win, Handler => $sub };
}

# Destroy all of a window's registered keymaps.
sub remove_keymap {
  my ($self, $name) = @_;
  # Messy pattern. Modifying arrays in for loops is... owch.
  my @new;
  foreach (@{ $self->{_Keymaps} }) {
    push @new, $_ unless $_->{Win}{Name} eq $name;
  }
  @{ $self->{_Keymaps} } = @new;
}

# Starts everything. The user can do this themselves, it doesn't matter, as long as POE
# runs.
sub start {
  $poe_kernel->run;
  exit;
}

# Do a blocking style "wait for input" without actually blocking.
sub wait_input {
  my $self = shift;
  $poe_kernel->run_one_timeslice until defined $self->{_BlockingInput};
  return delete $self->{_BlockingInput};
}

# Wrap a quick keymap to simply return input to wait_input(). Don't layer us -- one at a
# time please.
sub start_in {
  my $self = shift;
  $self->install_keymap({ Name => "Tmp" }, sub {
    my (undef, $key) = @_;
    # Since our window is bogus anyway ("Tmp"), use the closure magic stuff
    $self->{_BlockingInput} = $key;
  });
}

# Remove that temporary key returner.
sub stop_in {
  my $self = shift;
  $self->remove_keymap("Tmp");
}

# Caching color codes is highly useful in big maps. Believe me.
memoize("paint");

# Returns the attribute for the specified color.
sub paint {
	shift;	 # We don't need a reference to the singleton
  my ($fg, $bg);
  if (@_ == 1) {
    ($fg, $bg) = ($_[0]) ? (split("/", shift)) : ("grey", "black");
  } else {
    ($fg, $bg) = @_;
  }
  # If the foreground and background color are the same, something screws up.
  # So prevent that.
  $bg ||= "black";
  $fg = "black" if $bg =~ m/^(grey|gray|white)$/i and $fg =~ m/^(grey|gray|white)$/i;
  return _trans_color($fg, $bg);
}

# Ask Curses for actual low-level color value.
sub _trans_color {
  my ($fg, $bg) = @_;
  my $bright = 0;
  my %colors = (
    blue   => 5,
    red    => 2,
    green  => 3,
    aqua   => 7,
    cyan   => 7,
    purple => 6,
    orange => 4,
    yellow => 4,
    black  => 1,
    grey   => 8,
    gray   => 8,
    white  => 8
  );
  my $color = $colors{lc $fg};

  # Now we determine whether we're bright or dim.
  # If our first letter is capitalized, we're bright.
  $bright = 1 if substr($fg, 0, 1) eq uc substr($fg, 0, 1);
  # But then we have all these special cases. Simplify fg.
  $fg = lc $fg;
  $bright = 0 if $fg eq "orange";
  $bright = 1 if $fg eq "yellow";
  $bright = 0 if $fg eq "grey" or $fg eq "gray";
  $bright = 1 if $fg eq "white";

	# Don't forget background color. >_<
	$color += 8 * ($colors{lc $bg} - 1);

  my $attrib = $bright ? A_BOLD() : A_NORMAL();
  return $attrib | COLOR_PAIR($color);
}

42;
