package Object;

use strict;
use warnings;
use include;

use EventDB;

# Make a new template from stuff defined in a Perl module.
sub make {
  my ($class, $name, %data) = @_;

  my $template;
  # Either derive from something or be a root object.
  if (my $base = delete $data{From}) {
    $data{Base} = GAME->{Templates}{$base};
  }
  $template = bless {
    Stages => ["Before", "On", "After"],  # This is a "default Object property"
    Type   => $name,
    %data
  }, "Object";

  # Finally store the object
  GAME->{Templates}{$name} = $template;
}

# Make a new template based off another.
sub derive {
  my ($base, %dat) = @_;
  my $self = bless {
    Base => $base,
    %dat  # We gotta trust em to only send in class data
  }, "Object";

  return $self;
}

# Make a new instance object based off (hopefully) a template.
sub new {
  my ($base, %dat) = @_;
  my $self = bless {
    Base => $base,
  }, "Object";
  push @{ GAME->{Objects} }, $self;
  $self->{ID} = $#{ GAME->{Objects} };

  # Go through the constructor routines
  # Start with the "most specific" routine
  my $init = $self->g("Init");
  while (1) {
    last unless $init;
    # And continue going through constructor routines as long as they tell us
    # They have to explicitly tell us what Template's init routine to do next... normally we
    # don't want constructor chaining.
    my $code = $init->($self, \%dat);
    if ($code and $code ne "1") {
      # They told us what type to do next
      $init = GAME->{Templates}{$code}->g("Init");
    } else {
      last;
    }
  }

  # Make sure there's no remaining unprocessed data.
  if (%dat) {
    debug \%dat;
    die "didnt finish processing new object's data!";
  }

  return $self;
}

# Let an instance quickly initialize passed in data.
sub _construct {
  my ($self, $opts, @keys) = @_;
  foreach (@keys) {
    $self->{$_} = delete $opts->{$_} if defined $opts->{$_};
  }
}

# Install an event reaction.
sub react {
  my ($self, $stage, $verb, %dat) = @_;

  # Set a flag so we knows we can do this action.
  $self->{Actions}{$verb} = 1;

  GAME->{EventDB}->react(
    actor => $self,
    stage => $stage,
    verb  => $verb,
    %dat
  );
}

# Return the "most specific" set of this data. Assume everything's a plain hash -- no
# arrays, no nested Objects. Extremely simplified because this is often abused.
sub g {
  my ($self, $string) = @_;
  my @string = split(/\./, $string);
  my @bases = ($self, $self->bases);

  # Check each object, starting with the (maybe) instance
  BASE: foreach my $obj (@bases) {
    my $look = $obj;
    foreach my $key (@string) {
      $look = $look->{$key};
      next BASE unless defined $look;
    }
    return $look;   # The answer!
  }

  return; # Nobody had it
}

# Return a specific->general list of our base objects.
sub bases {
  my $self = shift;
  my @bases;
  my $base = $self;
  while ($base = $base->{Base}) {
    push @bases, $base;
  }
  return @bases;
}

# Here's the AUTOLOAD magic that calls methods or goes through the action sequence.
our $AUTOLOAD;

sub AUTOLOAD {
  my $self = shift;
  my $do = $AUTOLOAD;
  $do =~ s/^.*:://;
  return if $do eq "DESTROY";
  my $try = 1 if $do =~ s/^try_//;
  if ($self->g("Actions.$do")) {
    # TODO: The fact that we're TRYing to do it means it's an intermediate step in a
    # reaction.
    return $self->_do_act($do, @_);
  } elsif (my $method = $self->g("Methods.$do")) {
    return $method->($self, @_);
  } else {
    die "Object can't do $do\n";
  }
}

# Send all the signals and redirect actions and handle scheduling actions and all that
# stuff.
sub _do_act {
  my ($self, $action, @args) = @_;

  # Pass a heap around to let reactions modify the arguments and stuff
  my $heap = {                                                            
    Args => \@args   # AAAAAARRRGS... sorry.                         
  };

  my $return = 1;   # Unless we were explicitly :STOPed, assume success
  log_this($self->id . " is doing $action");
  log_push;
  foreach my $stage (@{ $self->g("Stages") }) {
    my @code = GAME->{EventDB}->signal(
      when   => $stage,
      action => $action,
      actor  => $self,
      heap   => $heap
    );
    $return = 0, last if $code[0] eq ":STOP";
    $return = $self->_do_act(@code), last if shift @code eq ":REDIRECT";
    # Otherwise, fine.
  }
  log_pop();
  return $return;
}

# Describe an event involving objects, replacing placeholders with properly formed names.
no warnings "redefine";   # I know i know i know!
sub msg {
  my $self = shift;
  my $msg = shift;

  my $custom = $self->g("UniquePossessive");  # Indicates a special form in general

  # 1) Possessives
  my @args = map {
    if ($custom) {
      $custom;
    } else {
      $_->specific_name . "'s";
    }
  } @_;
  $msg =~ s/\[(\d+)P\]/$args[$1 - 1]/g;

  # 2) Subject
  my $subjname = $self->specific_name;
  $msg =~ s/\[subj\]/$subjname/gi;

  # 3) Normal arguments
  @args = map { $_->specific_name } @_;
  $msg =~ s/\[(\d+)\]/$args[$1 - 1]/g;

  # 4) Regular verbs
  my $suffix = $custom ? "" : "s";
  $msg =~ s/\[(\w+)\]/$1$suffix/g;

  # 5) Irregular verbs
  if ($custom) {
    # First one in the verb list
    $msg =~ s/\[([\w']+)\/([\w']+)\]/$1/g;
  } else {
    # Second one
    $msg =~ s/\[([\w']+)\/([\w']+)\]/$2/g;
  }

  # 6) Capitalize first letter. But ignore leading <\w>s!
  if ($msg =~ m/^</) {
    $msg =~ s/^(<\w+>)(\w)/$1\U$2/;
  } else {
    $msg =~ s/^(\w)/\U$1/;
  }

  # 7) Capitalize first letter after every punctuation mark.
  $msg =~ s/(\.|!|\?) (\w)/$1 \U$2/g;

  # And finally say it!
  include::msg($msg);   # Call the real one
}
use warnings "redefine";

# Describe something with definite articles.
sub specific_name {
  my $self = shift;

  # Anything can be a unique.
  if (my $name = $self->g("UniqueName")) {
    return $name;   # No work to do
  }

  if ($self->{Qty} and $self->{Qty} > 1) {
    return "$self->{Qty} " . $self->name . "s";
  } else {
    return "the " . $self->name;
  }
}

# Describe something with indefinite articles.
sub general_name {
  my $self = shift;

  # Anything can be a unique.
  if (my $name = $self->g("UniqueName")) {
    return $name;   # No work to do
  }

  if ($self->{Qty} and $self->{Qty} > 1) {
    return "$self->{Qty} " . $self->name . "s";
  } else {
    my $name = $self->name;
    return $name =~ m/^(a|e|i|o|u)/ ? "an $name" : "a $name";
  }
}

# Identify this template or object uniquely.
sub id {
  my $self = shift;

  return $self->{Type} if $self->{Type};  # Template
  return $self->g("Type") . "_$self->{ID}";
}

42;
